//-----------------------------------------------------------------------------
// hbp_Comm.C -- Category of objects used for communication.  For example,
// Remote Method Invocation (RMI) and global pointer get/put.
//-----------------------------------------------------------------------------

#include "hbp_Comm.h"
#include "hpcxx_rtsImpl.h"
#include <iostream.h>


static const int CMD_BUF_SIZE = CACHE_SLINE_SIZE;


hbp_Comm::hbp_Comm(HPCxx_Mutex* mutex, hbp_state* hbp, unsigned int dst_pg,
	    unsigned int dst_off, unsigned int dst_port,
	    PageBuffer* sendBuf, PageBuffer* ackBuf)
         : HPCxx_Comm(mutex)

{
  this->hbp      = hbp;
  this->dst_pg   = dst_pg;
  this->dst_off  = dst_off;
  this->dst_port = dst_port;
  this->sendBuf  = sendBuf;
  this->ackBuf   = ackBuf;
  this->buffer   = sendBuf;
}


hbp_Comm::~hbp_Comm(void)
{
  delete sendBuf;
  delete ackBuf;
}


int hbp_Comm::messageID(void) const
{
  hbp_CmdHdr* hdr = (hbp_CmdHdr*) sendBuf_->buffer();
  return hdr->msgId;
}


int hbp_Comm::messageSender(void) const
{
  hbp_CmdHdr* hdr = (hbp_CmdHdr*) sendBuf_->buffer();
  return hdr->senderId;
}


int hbp_Comm::messageThread(void) const
{
  hbp_CmdHdr* hdr = (hbp_CmdHdr*) sendBuf_->buffer();
  return hdr->threadId;
}


int hbp_Comm::messageType(void) const
{
  hbp_CmdHdr* hdr = (hbp_CmdHdr*) sendBuf_->buffer();
  return hdr->msgType;
}


HPCxx_Buffer& hbp_Comm::messageBuffer(void) const
{
  return *sendBuf_;
}


void hbp_Comm::setMessageID(int msgId, bool unique)
{
  hbp_CmdHdr* hdr = (hbp_CmdHdr*) sendBuf_->buffer();
  hdr->msgId = msgId;
}


void hbp_Comm::setMessageType(int msgType)
{
  hbp_CmdHdr* hdr = (hbp_CmdHdr*) sendBuf_->buffer();
  hdr->msgType = msgType;
}


void hbp_Comm::setSender(const HPCxx_ContextID& sender)
{
  hbp_CmdHdr* hdr = (hbp_CmdHdr*) sendBuf_->buffer();
  hdr->senderId = sender;
}


void hbp_Comm::setSenderIndex(int index)
{
  hbp_CmdHdr* hdr = (hbp_CmdHdr*) sendBuf_->buffer();
  hdr->senderIndex = index;
}


void hbp_Comm::send(void)
{
  mutex_ -> lock();
  hbp_writec(hbp_, NULL, sendBuf_->size(),
	       sendBuf_->page(), sendBuf_->pageOffset(),
	       destination_, dst_port_, dst_pg_, dst_off_);
  mutex_ -> unlock();
}


void hbp_Comm::send(HPCxx_Buffer& buffer)
{
  if (buffer.impl_.size() > CMD_BUF_SIZE) return;

  mutex_ -> lock();
  hbp_writec(hbp_, buffer.impl_.buffer(), buffer.impl_size(),
	       sendBuf_->page(), sendBuf_->pageOffset(),
	       destination_, dst_port_, dst_pg_, dst_off_);
  mutex_ -> unlock();
}


void hbp_Comm::send(void* buf, size_t size)
{
  if (size > CMD_BUF_SIZE) return;

  mutex_ -> lock();
  hbp_writec(hbp_, buf, size,
	       sendBuf_->page(), sendBuf_->pageOffset(),
	       destination_, dst_port_, dst_pg_, dst_off_);
  MPI_Isend(buf, size, MPI_BYTE, destination_, tag_, comm_, &request_);
  mutex_ -> unlock();
}


size_t hbp_Comm::recv(void* buf, size_t size)
{
  // No lock needed here, just copy from pinned memory.  This assumes
  // that the ack has arrived. The arrival of a message must be
  // discovered by a poll of the underlying bps software layer.

  if (size > CMD_BUF_SIZE) size = CMD_BUF_SIZE;
  bcopy(ackBuf_->buffer(), buf, size);
  return size;
}

ostream& operator<<(ostream& s, const hbp_Comm& c)
{
  return s << "tag = " << (void*) c.messageTag() << ", sender = " << c.sender_
           << ", dest = " << c.destination_
           << ", size = " << c.messageBuffer().size()
           << ", type = " << c.messageType() << ", id = " << c.messageID();
//           << ", thread = " << c.messageThread();
}

